import inspect
from typing import Type
from unittest import TestCase, skip

from egon.decorators import as_node, as_source, as_target
from egon.nodes import Node, Source, Target


def function_add(x: int, y: int) -> int:
    """Return the sum of two numbers"""

    return x + y


class BaseTests:
    """Generic tests applicable to multiple function decorators"""

    wrapper: callable  # The function decorator to test
    return_type: Type  # The expected return type of the decorator

    @classmethod
    def setUpClass(cls) -> None:
        """Wrap the ``function_add`` function with a decorator"""

        cls.wrapped_function = cls.wrapper(function_add)

    def test_wrapped_return_type(self) -> None:
        """Test the function decorator returns the expected object type"""

        self.assertIsInstance(self.wrapped_function, self.return_type)

    def test_wrapped_is_still_callable(self) -> None:
        """Test wrapped functions are still callable"""

        test_args = 1, 2
        expected_return = function_add(*test_args)
        wrapped_return = self.wrapped_function(*test_args)
        self.assertEqual(expected_return, wrapped_return)

    def test_wrapped_function_maintains_signature(self) -> None:
        """Test the wrapped function has the same signature as the original function"""

        original_sig = inspect.getfullargspec(function_add)
        wrapped_sig = inspect.getfullargspec(self.wrapped_function)
        self.assertEqual(wrapped_sig, original_sig)


class WrappedAsSource(BaseTests, TestCase):
    """Tests for the ``as_source`` decorator"""

    wrapper = staticmethod(as_source)
    return_type = Source

    @skip
    def test_wrapped_is_still_callable(self) -> None:
        """Test wrapped functions are still callable"""

    def test_wrapped_is_still_generator(self) -> None:
        """Test the wrapped function still acts as a generator"""

        def generator():
            for i in range(10):
                yield i

        wrapped = self.wrapper(generator)
        self.assertListEqual(list(generator()), list(wrapped()))


class WrappedAsNode(BaseTests, TestCase):
    """Tests for the ``as_node`` decorator"""

    wrapper = staticmethod(as_node)
    return_type = Node


class WrappedAsTarget(BaseTests, TestCase):
    """Tests for the ``as_target`` decorator"""

    wrapper = staticmethod(as_target)
    return_type = Target



"""Function decorators from the ``decorators`` module provide a shorthand for
creating pipeline nodes from pre-built functions. Different decorators
produce different kinds of pipeline nodes.
"""

import inspect
from typing import Any, Callable, Generator

from boltons.funcutils import wraps

from . import connectors, nodes

GeneratorFunction = Callable[[], Generator]


def _as_single_arg_func(func: callable) -> callable:
    """Wrap a callable so that it accepts at most one argument

    Args:
        func: The callable to wrap

    Returns:
        The wrapped function
    """

    if len(inspect.getfullargspec(func).args) <= 1:
        return func

    return lambda args: func(*args)


class WrappedSource(nodes.Source):
    """Wrapped callable as a Source-like object"""

    def __init__(self, func: GeneratorFunction) -> None:
        self.output = connectors.Output()
        self._func = _as_single_arg_func(func)
        super().__init__()

    def action(self) -> None:
        """Call the wrapped generator and load results into the ``output`` connector"""

        for x in self._func():
            self.output.put(x)

    def __str__(self) -> str:  # pragma: no cover
        return f'<WrappedSource(wrapped_function={self._func.__name__}) object at {hex(id(self))}>'


class WrappedTarget(nodes.Target):
    """Wrapped callable as a Target-like object"""

    def __init__(self, func: callable) -> None:
        self.input = connectors.Input()
        self._func = _as_single_arg_func(func)
        super().__init__()

    def action(self) -> None:
        """Call the wrapped function using data from the ``input`` connector"""

        for data in self.input.iter_get():
            self._func(data)

    def __str__(self) -> str:  # pragma: no cover
        return f'<WrappedTarget(wrapped_function={self._func.__name__}) object at {hex(id(self))}>'


class WrappedNode(nodes.Node):
    """Wrapped callable as a Node-like object"""

    def __init__(self, func: callable) -> None:
        self.input = connectors.Input()
        self.output = connectors.Output()
        self._func = _as_single_arg_func(func)
        super().__init__()

    def action(self) -> None:
        """Call the wrapped function and put it's return in the ``output`` connector."""

        for data in self.input.iter_get():
            self.output.put(self._func(data))

    def __str__(self) -> str:  # pragma: no cover
        return f'<WrappedNode(wrapped_function={self._func.__name__}) object at {hex(id(self))}>'


def as_source(func: GeneratorFunction) -> nodes.Source:
    """Decorator for wrapping a callable as a pipeline ``Source`` object"""

    class Wrapper(WrappedSource):
        @staticmethod
        @wraps(func)
        def __call__() -> Any:
            return func()

    return Wrapper(func)


def as_target(func: callable) -> nodes.Target:
    """Decorator for wrapping a callable as a pipeline ``Target`` object"""

    class Wrapper(WrappedTarget):
        @staticmethod
        @wraps(func)
        def __call__(*args, **kwargs) -> Any:
            return func(*args, **kwargs)

    return Wrapper(func)


def as_node(func: callable) -> nodes.Node:
    """Decorator for wrapping a callable as a pipeline ``Node`` object"""

    class Wrapper(WrappedNode):
        @staticmethod
        @wraps(func)
        def __call__(*args, **kwargs) -> Any:
            return func(*args, **kwargs)

    return Wrapper(func)



from unittest import TestCase

from tests.mock import MockSource


class OutputSet(TestCase):
    """Test data storage in ``Output`` instances"""

    def setUp(self) -> None:
        """Define an ``Input`` instance"""

        # Create a node with an output connector
        self.source = MockSource()

    def test_stores_value_in_queue(self) -> None:
        """Test the ``put`` method retrieves data from the underlying queue"""

        test_val = 'test_val'
        self.source.output.put(test_val)
        self.assertEqual(self.source.output._queue.get(), test_val)



from functools import partial
from unittest import TestCase

from . import mock


class ProcessAllocation(TestCase):
    """Test ``Node`` instances fork the correct number of processes"""

    def test_allocation_at_init(self) -> None:
        """Test the correct number of processes are allocated at init"""

        num_processes = 4
        node = mock.MockNode(num_processes)
        self.assertEqual(num_processes, node.num_processes)
        self.assertEqual(num_processes, len(node._processes))

    def test_reallocation(self) -> None:
        """Test a new set of processes are allocated when the number of processes is changed"""

        num_processes = 4
        node = mock.MockNode(num_processes=1)
        node.num_processes = 4
        self.assertEqual(num_processes, node.num_processes)

    def test_error_if_processes_are_alive(self) -> None:
        """Test a RuntimeError is raised when trying to realocate processes on a running node"""

        node = mock.MockNode()
        node._processes[0].start()
        with self.assertRaises(RuntimeError):
            node.num_processes = 1


class Execution(TestCase):
    """Test the execution of tasks assigned to a Node instance"""

    def setUp(self) -> None:
        """Create a testing node that tracks the execution method of it's methods"""

        self.node = mock.MockNode()

        # Track the call order of node functions
        self.call_list = []
        self.node.setup = partial(self.call_list.append, 'setup')
        self.node.action = partial(self.call_list.append, 'action')
        self.node.teardown = partial(self.call_list.append, 'teardown')

    def test_call_order(self) -> None:
        """Test that setup and teardown actions are called in the correct order"""

        expected_order = ['setup', 'action', 'teardown']
        self.node.execute()
        self.assertListEqual(self.call_list, expected_order)

    def test_process_is_finished_on_execute(self) -> None:
        """Test the ``process_finished`` property is updated after node execution"""

        self.assertFalse(self.node.process_finished, 'Default finished state is not False.')
        self.node.execute()
        self.assertTrue(self.node.process_finished)

    def test_node_is_finished_on_execute(self) -> None:
        """Test the ``node_finished`` property is updated after node execution"""

        self.assertFalse(self.node.node_finished, 'Default finished state is not False.')
        self.node._processes[0].start()
        self.node._processes[0].join()
        self.assertTrue(self.node.node_finished)


class TreeNavigation(TestCase):
    """Test ``Node`` instances are aware of their neighbors"""

    def setUp(self) -> None:
        """Create a tree of ``MockNode`` instances"""

        self.root = mock.MockSource()
        self.internal_node = mock.MockNode()
        self.leaf = mock.MockTarget()

        self.root.output.connect(self.internal_node.input)
        self.internal_node.output.connect(self.leaf.input)

    def test_upstream_nodes(self) -> None:
        """Test the inline node resolves the correct parent node"""

        self.assertEqual(self.root, self.internal_node.upstream_nodes()[0])

    def test_downstream_nodes(self) -> None:
        """Test the inline node resolves the correct child node"""

        self.assertEqual(self.leaf, self.internal_node.downstream_nodes()[0])


class ExpectingData(TestCase):
    """Tests for the ``expecting_data`` function

    The ``expecting_data`` function combines two booleans.
    This class evaluates all four suqares of the corresponding truth table
    """

    def setUp(self) -> None:
        """Create a tree of ``MockNode`` instances"""

        # Fork zero processes so we can control the node finished state as the state of the daemon process
        self.root = mock.MockSource(num_processes=0)
        self.node = mock.MockNode(num_processes=0)
        self.root.output.connect(self.node.input)

    def test_false_for_empty_queue_and_finished_parent(self) -> None:
        """Test the return is False for a EMPTY queue and a FINISHED PARENT node"""

        self.root.process_finished = True
        self.assertFalse(self.node.expecting_data())

    def test_true_if_input_queue_has_data(self) -> None:
        """Test the return is True for a NOT EMPTY queue and a FINISHED PARENT node"""

        self.root.process_finished = True
        self.node.input._queue.put(5)
        self.assertTrue(self.node.expecting_data())

    def test_true_if_parent_is_running(self) -> None:
        """Test the return is True for a EMPTY queue and a NOT FINISHED PARENT node"""

        self.root.process_finished = False
        self.assertTrue(self.node.expecting_data())

    def test_true_if_input_queue_has_data_and_parent_is_running(self) -> None:
        """Test the return is True for a NOT EMPTY queue and a NOT FINISHED PARENT node"""

        self.root.process_finished = False
        self.node.input._queue.put(5)
        self.assertTrue(self.node.expecting_data())



"""The ``exceptions`` class defines custom exceptions raised by the ``egon`` package."""


class MissingConnectionError(Exception):
    """Raised when a connector is left unconnected at pipeline runtime."""


class MalformedSourceError(Exception):
    """Raised when a ``Source`` object is created with the wrong type of connectors."""


class MalformedTargetError(Exception):
    """Raised when a ``Target`` object is created with the wrong type of connectors."""


class OrphanedNodeError(Exception):
    """Raised when a ``Node`` is inaccessible by the pipeline due to missing connectors."""


class OverwriteConnectionError(Exception):
    """Raised when trying to overwrite an existing connection between two connectors."""



"""Tests for connector objects defined in the ``connectors`` module"""

from unittest import TestCase

from egon import exceptions
from egon.connectors import Connector, Input, Output
from tests.mock import MockSource, MockTarget


class InstanceConnections(TestCase):
    """Test the connection of generic connector objects to other"""

    def setUp(self) -> None:
        """Define a generic ``Connector`` instance"""

        self.connector = Connector()

    def test_overwrite_error_on_connection_overwrite(self) -> None:
        """An error is raised when trying to overwrite an existing connection"""

        self.connector.connect(Input())
        with self.assertRaises(exceptions.OverwriteConnectionError):
            self.connector.connect(Input())

    def test_overwrite_error_on_connected_argument(self) -> None:
        """An error is raised when trying to connect to a connector with an established connection"""

        input_connector = Input()
        output_connector = Output()
        input_connector.connect(output_connector)

        with self.assertRaises(exceptions.OverwriteConnectionError):
            self.connector.connect(input_connector)

    def test_connected_instances_share_queue(self) -> None:
        """Test two connected instances share the same memory queue"""

        input_connector = Input()
        output_connector = Output()
        input_connector.connect(output_connector)
        self.assertIs(input_connector._queue, output_connector._queue)

    def test_is_connected_boolean(self) -> None:
        """The ``is_connected`` method returns the current connection state"""

        input_connector = Input()
        self.assertFalse(input_connector.is_connected())
        input_connector.connect(Output())
        self.assertTrue(input_connector.is_connected())

    def test_error_on_connection_to_same_type(self) -> None:
        """An error is raised when connecting two inputs together"""

        with self.assertRaises(ValueError):
            Connector().connect(Connector())


class PartnerMapping(TestCase):
    """Test connectors with an established connection correctly map to neighboring connectors/nodes"""

    def setUp(self) -> None:
        """Create two connected pipeline elements"""

        self.target = MockTarget()
        self.input_connector = self.target.input
        self.output_connector = MockSource().output
        self.output_connector.connect(self.input_connector)

    def test_is_aware_of_partner(self) -> None:
        """Test connectors map to the correct partner connector"""

        self.assertIs(self.input_connector.partner, self.output_connector)

    def test_is_aware_of_parent(self) -> None:
        """Test connectors map to their partner"""

        self.assertIs(self.input_connector.parent_node, self.target)


class InstanceDisconnect(TestCase):
    """Test the disconnection of two connectors"""

    def setUp(self) -> None:
        """Define an ``Input`` instance"""

        self.input = Input()
        self.output = Output()
        self.input.connect(self.output)

    def test_queue_is_rewritten(self) -> None:
        """Test connectors revert to having individual queues"""

        original_queue = self.input._queue
        self.input.disconnect()
        self.assertIsNot(self.input._queue, original_queue)
        self.assertIsNot(self.output._queue, original_queue)
        self.assertIsNot(self.input._queue, self.output._queue)

    def test_both_connectors_are_disconnected(self) -> None:
        """Test calling disconnect from one connector results in both connectors being disconnected"""

        self.input.disconnect()
        self.assertFalse(self.input.is_connected())
        self.assertFalse(self.output.is_connected())

    @staticmethod
    def test_no_error_on_successive_disconnect() -> None:
        """Test no errors are raised when disconnecting an instance with no connection"""

        Input().disconnect()



"""The Pipeline module defines the ``Pipeline`` class which is responsible
for representing collections of interconnected analysis steps (``Node``
objects) as a  single, coherent analysis pipeline. ``Pipeline`` instances
are also responsible for starting/stopping forked processes and validating
that nodes are properly interconnected.
"""

from __future__ import annotations

from asyncio.subprocess import Process
from inspect import getmembers
from typing import List

from . import nodes
from .nodes import Node


class Pipeline:
    """Manages a collection of nodes as a single analysis pipeline"""

    def validate(self) -> None:
        """Set up the pipeline and check for any invalide node states"""

        # Make sure the nodes are in a runnable condition before we start spawning _processes
        for node in self.get_nodes():
            node.validate()

    def _get_processes(self) -> List[Process]:
        """Return a list of processes forked by pipeline nodes"""

        # Collect all of the processes assigned to each node
        processes = []
        for node in self.get_nodes():
            processes.extend(node._processes)

        return processes

    def get_nodes(self) -> List[Node]:
        """Return a list of all nodes in the pipeline"""

        return [getattr(self, a[0]) for a in getmembers(self, lambda a: isinstance(a, nodes.AbstractNode))]

    def num_processes(self) -> int:
        """The number of processes forked by to the pipeline"""

        return len(self._get_processes())

    def kill(self) -> None:
        """Kill all running pipeline processes without trying to exit gracefully"""

        for p in self._get_processes():
            p.terminate()

    def run(self) -> None:
        """Start all pipeline processes and block execution until all processes exit"""

        self.run_async()
        self.wait_for_exit()

    def wait_for_exit(self) -> None:
        """Wait for the pipeline to finish running before continuing execution"""

        for p in self._get_processes():
            p.join()

    def run_async(self) -> None:
        """Start all processes asynchronously"""

        for p in self._get_processes():
            p.start()



"""The ``nodes`` module supports the construction of individual pipeline nodes.
``Source``, ``Node``,  and ``Target`` classes are provided for creating nodes
that produce, analyze, and costume data respectively.
"""

from __future__ import annotations

import abc
import inspect
import multiprocessing as mp
from abc import ABC
from typing import List, Union, Collection

from . import connectors, exceptions


def _get_nodes_from_connectors(connector_list: Collection[connectors.Connector]) -> List:
    """Return the parent nodes from a list of ``Connector`` objects
    
    Args:
        connector_list: The connectors to get parents of
        
    Returns:
        A list of node instances
    """

    nodes = []
    for c in connector_list:
        if c.is_connected():
            nodes.append(c.partner.parent_node)

    return nodes


class AbstractNode(abc.ABC):
    """Base class for constructing pipeline nodes"""

    def __init__(self, num_processes=1) -> None:
        """Represents a single pipeline node"""

        # Note that we use the memory address and not the ``pid`` attribute.
        # ``pid`` is only set after the process is started
        self._processes = [mp.Process(target=self.execute) for _ in range(num_processes)]
        self._states = mp.Manager().dict({id(p): False for p in self._processes})

        self._current_process_state = False
        for connection in self._get_attrs(connectors.Connector):
            connection._node = self

    @property
    def num_processes(self) -> int:
        """The number of processes assigned to the current node"""

        return len(self._processes)

    @num_processes.setter
    def num_processes(self, num_processes) -> None:
        """The number of processes assigned to the current node"""

        if any(p.is_alive() for p in self._processes):
            raise RuntimeError('Cannot change number of processes while node is running.')

        if self.num_processes == num_processes:
            return

        self._processes = [mp.Process(target=self.execute) for _ in range(num_processes)]
        self._states = mp.Manager().dict({id(p): False for p in self._processes})

    @property
    def process_finished(self) -> bool:
        """Return whether the current process has finished processing data"""

        # Use get in case called from a process not forked by the class __init__
        return self._states.get(mp.current_process().pid, self._current_process_state)

    @process_finished.setter
    def process_finished(self, state: bool) -> None:
        self._states[id(mp.current_process())] = self._current_process_state = state

    @property
    def node_finished(self) -> bool:
        """Return whether all node processes have finished processing data"""

        return all(self._states.values())

    @abc.abstractmethod
    def validate(self) -> None:
        """Raise an exception if the node object was constructed improperly

        Raises:
            ValueError: For an invalid instance construction
        """

    def _validate_connections(self) -> None:
        """Raise an exception if any of the node's Inputs/Outputs are missing connections

        Raises:
            MissingConnectionError: For an invalid instance construction
        """

        for conn in self._get_attrs(connectors.Connector):
            if not conn.is_connected():
                raise exceptions.MissingConnectionError(
                    f'Connector {conn} does not have an established connection (Node: {conn.parent_node})')

    def _get_attrs(self, attr_type=None) -> List:
        """Return a list of instance attributes matching the given type

        Args:
            attr_type: The object type to search for

        Returns:
            A list of attributes of type ``attr_type``
        """

        return [getattr(self, a[0]) for a in inspect.getmembers(self, lambda a: isinstance(a, attr_type))]

    def upstream_nodes(self) -> List[Union[Source, Node]]:
        """Returns a list of nodes that are upstream from the current node"""

        return _get_nodes_from_connectors(self._get_attrs(connectors.Input))

    def downstream_nodes(self) -> List[Union[Node, Target]]:
        """Returns a list of nodes that are downstream from the current node"""

        return _get_nodes_from_connectors(self._get_attrs(connectors.Output))

    def setup(self) -> None:
        """Setup tasks called before running ``action``"""

    @abc.abstractmethod
    def action(self) -> None:
        """The primary analysis task performed by this node"""

    def teardown(self) -> None:
        """Teardown tasks called after running ``action``"""

    def execute(self) -> None:
        """Execute the pipeline node

        Execution includes all ``setup``, ``action``, and ``teardown`` tasks.
        """

        self.setup()
        self.action()
        self.teardown()
        self.process_finished = True

    def expecting_data(self) -> bool:
        """Return whether the node is still expecting data from upstream"""

        for input_connector in self._get_attrs(connectors.Input):
            if not (input_connector.empty() and input_connector.partner.parent_node.node_finished):
                return True

        return False


class Source(AbstractNode, ABC):
    """A pipeline process that only has output streams"""

    def validate(self) -> None:
        """Raise exception if the object is not a valid instance

        Raises:
            MalformedSourceError: For an invalid instance construction
            OrphanedNodeError: For an instance that is inaccessible by connectors
        """

        if self._get_attrs(connectors.Input):
            raise exceptions.MalformedSourceError('Source objects cannot have upstream components.')

        if not self._get_attrs(connectors.Output):
            raise exceptions.OrphanedNodeError('Source has no output connectors and is inaccessible by the pipeline.')

        self._validate_connections()


class Target(AbstractNode, ABC):
    """A pipeline process that only has input streams"""

    def validate(self) -> None:
        """Raise exception if the object is not a valid instance

        Raises:
            MalformedTargetError: For an invalid instance construction
            OrphanedNodeError: For an instance that is inaccessible by connectors
        """

        if self._get_attrs(connectors.Output):
            raise exceptions.MalformedTargetError('Source objects cannot have upstream components.')

        if not self._get_attrs(connectors.Input):
            raise exceptions.OrphanedNodeError('Target has no input connectors and is inaccessible by the pipeline.')

        self._validate_connections()


class Node(Target, Source, ABC):
    """A pipeline process that can have any number of input or output streams"""

    def validate(self) -> None:
        """Raise exception if the object is not a valid instance

        Raises:
            OrphanedNodeError: For an instance that is inaccessible by connectors
        """

        if not self._get_attrs(connectors.Connector):
            raise exceptions.OrphanedNodeError('Node has no associated connectors and is inaccessible by the pipeline.')

        self._validate_connections()



from asyncio import sleep

from egon import nodes
from egon.connectors import Input, Output
from egon.pipeline import Pipeline


class MockSource(nodes.Source):
    """A ``Source`` subclass that implements placeholder functions for abstract methods"""

    def __init__(self, num_processes=1) -> None:
        self.output = Output()
        super(MockSource, self).__init__(num_processes)

    def action(self) -> None:
        """Placeholder function to satisfy requirements of abstract parent"""

        sleep(10)


class MockTarget(nodes.Target):
    """A ``Target`` subclass that implements placeholder functions for abstract methods"""

    def __init__(self, num_processes=1) -> None:
        self.input = Input()
        super(MockTarget, self).__init__(num_processes)

    def action(self) -> None:
        """Placeholder function to satisfy requirements of abstract parent"""

        sleep(100)


class MockNode(nodes.Node):
    """A ``Node`` subclass that implements placeholder functions for abstract methods"""

    def __init__(self, num_processes=1) -> None:
        self.output = Output()
        self.input = Input()
        super(MockNode, self).__init__(num_processes)

    def action(self) -> None:
        """Placeholder function to satisfy requirements of abstract parent"""

        sleep(10)


class MockPipeline(Pipeline):
    """A mock pipeline with a root and a leaf"""

    def __init__(self) -> None:
        self.root = MockSource(num_processes=2)
        self.leaf = MockTarget()
        self.root.output.connect(self.leaf.input)

        self.validate()

    def all_alive(self) -> bool:
        """Return if all processes managed by the pipeline are alive"""

        return all(p.is_alive() for p in self._get_processes())

    def any_alive(self) -> bool:
        """Return if any process managed by the pipeline are alive"""

        return any(p.is_alive() for p in self._get_processes())



from unittest import TestCase

from egon.connectors import KillSignal
from tests.mock import MockTarget, MockSource


class InputGet(TestCase):
    """Test data retrieval from ``Input`` instances"""

    def setUp(self) -> None:
        """Define a node with an attached ``Input`` instance"""

        self.target = MockTarget(num_processes=0)  # Run node in current process only

    def test_error_on_non_positive_refresh(self) -> None:
        """Test a ValueError is raised when ``refresh_interval`` is not a positive number"""

        with self.assertRaises(ValueError):
            self.target.input.get(timeout=15, refresh_interval=0)

        with self.assertRaises(ValueError):
            self.target.input.get(timeout=15, refresh_interval=-1)

    def test_returns_queue_value(self) -> None:
        """Test the ``get`` method retrieves data from the underlying queue"""

        test_val = 'test_val'
        self.target.input._queue.put(test_val)
        self.assertEqual(self.target.input.get(timeout=1000), test_val)

    def test_kill_signal_on_finished_parent_node(self) -> None:
        """Test a kill signal is returned if the parent node if finished"""

        source = MockSource(num_processes=0)
        source.output.connect(self.target.input)
        source.process_finished = True
        self.assertFalse(self.target.expecting_data())
        self.assertIs(self.target.input.get(timeout=15), KillSignal)

    def test_timeout_raises_timeout_error(self) -> None:
        """Test a ``TimeoutError`` is raise on timeout"""

        with self.assertRaises(TimeoutError):
            self.target.input.get(timeout=1)


class InputIterGet(TestCase):
    """Test iteration behavior of the ``iter_get`` method"""

    def setUp(self) -> None:
        """Define a node with an attached ``Input`` instance"""

        # Create a node with an input connector
        self.target = MockTarget()

    def test_raises_stop_iteration_on_kill_signal(self) -> None:
        """Test the iterator exits once it reaches a KillSignal object"""

        self.target.input._queue.put(KillSignal)
        with self.assertRaises(StopIteration):
            next(self.target.input.iter_get())

    def test_returns_queue_value(self) -> None:
        """Test the ``get`` method retrieves data from the underlying queue"""

        test_val = 'test_val'
        self.target.input._queue.put(test_val)
        self.assertEqual(next(self.target.input.iter_get()), test_val)



"""Tests for the ``Pipeline`` class"""

from time import sleep
from unittest import TestCase

from egon.connectors import Output
from egon.exceptions import OrphanedNodeError, MissingConnectionError
from egon.nodes import Node
from egon.pipeline import Pipeline
from tests.mock import MockSource, MockNode, MockPipeline


class StartStopCommands(TestCase):
    """Test processes are launched and terminated on command"""

    def runTest(self) -> None:
        """Launch and then kill the process manager"""

        pipeline = MockPipeline()
        self.assertFalse(pipeline.any_alive())

        pipeline.run_async()
        self.assertTrue(pipeline.all_alive())

        pipeline.kill()
        sleep(1)  # Give the process time to exit
        self.assertFalse(pipeline.any_alive())


class ProcessDiscovery(TestCase):
    """Test the pipeline is aware of all processes forked by it's nodes"""

    def test_collected_processes_match_nodes(self) -> None:
        """Test ``_get_processes`` returns forked processes from all pipeline nodes"""

        pipeline = MockPipeline()
        expected_processes = []
        expected_processes.extend(pipeline.root._processes)
        expected_processes.extend(pipeline.leaf._processes)
        self.assertCountEqual(expected_processes, pipeline._get_processes())

    def test_process_count(self) -> None:
        """Test the pipelines process count matches the sum of processes allocated to each node"""

        pipeline = MockPipeline()
        expected_count = pipeline.root.num_processes + pipeline.leaf.num_processes
        self.assertEqual(expected_count, pipeline.num_processes())


class PipelineValidation(TestCase):
    """Test appropriate errors are raised for an invalid pipeline."""

    def test_orphaned_node(self) -> None:
        """Test a ``OrphanedNodeError`` for an unreachable node"""

        class OrphanedNode(Node):
            def action(self) -> None:
                pass

        class Pipe(Pipeline):

            def __init__(self) -> None:
                self.node = OrphanedNode()

        with self.assertRaises(OrphanedNodeError):
            Pipe().validate()

    def test_missing_connection(self) -> None:
        """Test a ``MissingConnectionError`` for an unconnected connector"""

        class Pipe(Pipeline):
            def __init__(self) -> None:
                self.root = MockSource()
                self.node = MockNode()
                self.node.second_output = Output()
                self.root.output.connect(self.node.input)

        with self.assertRaises(MissingConnectionError):
            Pipe().validate()


class NodeDiscovery(TestCase):
    """Test the pipeline is aware of all of it's nodes"""

    def runTest(self) -> None:
        pipeline = MockPipeline()
        expected_nodes = [pipeline.root, pipeline.leaf]
        recovered_nodes = pipeline.get_nodes()
        self.assertCountEqual(expected_nodes, recovered_nodes)



"""Tests for the ``DataStore`` class"""

import time
from unittest import TestCase

from egon.connectors import DataStore


class QueueProperties(TestCase):
    """Test  test the exposure of queue properties by the overlying ``DataStore`` class"""

    def setUp(self) -> None:
        """Create a ``DataStore`` instance"""

        self.data_store = DataStore(maxsize=1)

    def test_size_matches_queue_size(self) -> None:
        """Test the ``size`` method returns the size of the queue`"""

        self.assertEqual(self.data_store.size(), 0)
        self.data_store._queue.put(1)
        self.assertEqual(self.data_store.size(), 1)

    def test_full_state(self) -> None:
        """Test the ``full`` method returns the state of the queue"""

        self.assertFalse(self.data_store.full())
        self.data_store._queue.put(1)
        self.assertTrue(self.data_store.full())

    def test_empty_state(self) -> None:
        """Test the ``empty`` method returns the state of the queue"""

        self.assertTrue(self.data_store.empty())
        self.data_store._queue.put(1)

        # The value of Queue.empty() updates asynchronously
        time.sleep(1)

        self.assertFalse(self.data_store.empty())



from unittest import TestCase, skip

from egon import exceptions
from egon.connectors import Input, Output
from egon.nodes import Node, Source, Target


class BaseTests:
    """Generic tests applicable to multiple Node subclasses"""

    node_type = None  # The Node subclass type to test
    malformed_exception = None  # Validation error expected if the Node is malformed

    def setUp(self) -> None:
        """Create a subclass of the type being tested that implements abstract methods"""

        class TestSource(self.node_type):
            def action(self) -> None:
                pass

        self.test_class = TestSource()

    def test_error_if_orphaned(self) -> None:
        """Test ``OrphanedNodeError`` is raised for a Node with no connectors"""

        with self.assertRaises(exceptions.OrphanedNodeError):
            self.test_class.validate()

    def test_error_if_malformed(self) -> None:
        """Test a malformed error is raised"""

        self.test_class.input = Input()
        self.test_class.output = Output()
        with self.assertRaises(self.malformed_exception):
            self.test_class.validate()


class SourceValidation(BaseTests, TestCase):
    """Test the validation of ``Source`` nodes"""

    node_type = Source
    malformed_exception = exceptions.MalformedSourceError


class TargetValidation(BaseTests, TestCase):
    """Test the validation of ``Target`` nodes"""

    node_type = Target
    malformed_exception = exceptions.MalformedTargetError


class InlineValidation(BaseTests, TestCase):
    """Test the validation of ``Inline`` nodes"""

    node_type = Node

    @skip('Inline nodes cannot be malformed')
    def test_error_if_malformed(self) -> None:
        """Inline nodes cannot be malformed"""



"""``Connector`` objects are responsible for the communication of information
between nodes. Each connector instance is assigned to a single node and can be
used to send or receive data depending on the type of connector.
``Output`` connectors are used to send data and ``Input`` objects are
used to receive data.
"""

from __future__ import annotations

import multiprocessing as mp
from queue import Empty
from typing import Any, Optional, TYPE_CHECKING

from . import exceptions

if TYPE_CHECKING:  # pragma: no cover
    from .nodes import AbstractNode


class KillSignal:
    """Used to indicate that a process should exit"""


class DataStore:
    """Exposes select functionality from an underlying ``Queue`` object"""

    def __init__(self, maxsize=0) -> None:
        """Queue-like object for passing data between nodes and / or parallel processes

        Args:
            maxsize: Maximum number of objects to store in memory
        """

        self._queue = mp.Queue(maxsize=maxsize)

    def empty(self) -> bool:
        """Return if the connection queue is empty"""

        return self._queue.empty()

    def full(self) -> bool:
        """Return if the connection queue is full"""

        return self._queue.full()

    def size(self) -> int:
        """Return the size of the connection queue"""

        return self._queue.qsize()


class Connector(DataStore):
    """Base class for signal/slot-like connector objects"""

    def __init__(self, name: str = None) -> None:
        """Handles the communication of input/output data between pipeline nodes

        Args:
            name: Human readable name for the connector. Used in error and status messages.
        """

        super().__init__()
        self.name = name
        self._node: Optional[AbstractNode] = None  # The node that this connector is assigned to
        self._connected_partner: Optional[Connector] = None  # The connector object of another node

    def is_connected(self) -> bool:
        """Return whether the connector has an established connection"""

        return not (self._connected_partner is None)

    def connect(self, connector: Connector, maxsize: int = 0) -> None:
        """Establish the flow of data between this connector and another connector

        Args:
            connector: The connector object ot connect with
            maxsize: The maximum number of communicated items to store in memory
        """

        if type(connector) is type(self):
            raise ValueError('Cannot join together two connection objects of the same type.')

        if connector.is_connected() or self.is_connected():
            raise exceptions.OverwriteConnectionError(
                'Connector object already has a pre-established connection.')

        # Once a connection is established between two connectors, they share an internal queue
        self._connected_partner = connector
        connector._connected_partner = self
        connector._queue = self._queue = mp.Queue(maxsize)

    def disconnect(self) -> None:
        """Disconnect any established connections"""

        old_partner = self._connected_partner
        self._connected_partner = None
        self._queue = mp.Queue()

        if old_partner:
            old_partner.disconnect()

    @property
    def partner(self) -> Connector:
        """The connector object connected to this instance

        Returns ``None`` if no connection has been established
        """

        return self._connected_partner

    @property
    def parent_node(self) -> AbstractNode:
        """The parent node this connector is assigned to"""

        return self._node

    def __repr__(self) -> str:  # pragma: no cover
        return f'<egon.connectors.Input(name={self.name}) object at {hex(id(self))}>'


class Input(Connector):
    """Handles the input of data into a pipeline node"""

    def get(self, timeout: Optional[int] = None, refresh_interval: int = 2):
        """Blocking call to retrieve input data

        Releases automatically when no more data is coming from upstream

        Args:
            timeout: Raise a TimeoutError if data is not retrieved within the given number of seconds
            refresh_interval: How often to check if data is expected from upstream

        Raises:
            TimeOutError: Raised if the get call times out
        """

        if not refresh_interval > 0:
            raise ValueError('Connector refresh interval must be greater than zero.')

        timeout = timeout or float('inf')
        while timeout > 0:
            if self.is_connected() and not self.parent_node.expecting_data():
                return KillSignal

            try:
                return self._queue.get(timeout=min(timeout, refresh_interval))

            except (Empty, TimeoutError):
                timeout -= refresh_interval

        raise TimeoutError

    def iter_get(self) -> Any:
        """Iterator that returns input data

        Automatically exits once no more data is expected from upstream nodes.
        """

        while self.parent_node.expecting_data():
            data = self.get()
            if data is KillSignal:
                return

            yield data


class Output(Connector):
    """Handles the output of data from a pipeline node"""

    def put(self, x: Any) -> None:
        """Add data into the connector"""

        self._queue.put(x)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

from setuptools import find_packages, setup

import egon

# Get list of requirements
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(name='egon',
      version=egon.__version__,
      packages=find_packages(),
      keywords='Data Pipeline Parallel Pipeline',
      description='A lightweight, minimalist framework for the rapid development of analysis pipelines.',
      long_description=egon.__doc__,
      long_description_content_type='text/markdown',
      classifiers=[
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python'],
      author='Daniel Perrefort',
      author_email='djperrefort@pitt.edu',
      license='GPL v3',
      python_requires='>=3.6',
      install_requires=requirements
      )



"""Build a simple two node pipeline and test all input data makes it through to the end"""

from multiprocessing import Queue
from unittest import TestCase

from egon.decorators import as_source, as_target
from egon.pipeline import Pipeline

test_vals = list(range(10))  # Input values for the pipeline
queue = Queue()  # For storing pipeline outputs


@as_source
def sending_node() -> None:
    """Load data into the pipeline"""

    for i in test_vals:
        yield i


@as_target
def receiving_node(x) -> None:
    """Retrieve data out of the pipeline"""

    queue.put(x)


class AddingPipeline(Pipeline):
    """A pipeline for generating and then adding numbers"""

    def __init__(self) -> None:
        self.send_node = sending_node
        self.receive_node = receiving_node
        self.send_node.output.connect(self.receive_node.input)
        self.validate()


class TestPipelineThroughput(TestCase):
    """Test all data makes it through the pipeline"""

    def runTest(self) -> None:
        """Compare input and ouput pipeline values"""

        # Run should populate the global queue
        AddingPipeline().run()

        # Convert the queue into a list
        l = []
        while queue.qsize() != 0:
            l.append(queue.get())

        self.assertListEqual(test_vals, l)
